
Inductive env_skel (A B : Set) : Set :=
  EmptyE : env_skel A B
| ConsE  : A -> B -> env_skel A B -> env_skel A B .


Inductive env_lookup (A B : Set) : env_skel A B -> A -> B -> Prop :=
  LookupTop : 	forall a : A, 
 		forall b : B, 	
		forall e : env_skel A B, 
			env_lookup A B (ConsE A B a b e) a b

| LookupRec : 	forall a c : A,
		forall b d : B,
		forall e : env_skel A B,
			env_lookup A B e a b -> env_lookup A B (ConsE A B c d e) a b. 
 
Inductive stream_skel (A : Set) : Set :=
  EmptyS : stream_skel A
| ConsS  : A -> stream_skel A.

(* types *)

Inductive keyname : Set :=
 KN : nat -> keyname.

Inductive conflvl : Set :=
  Secret : conflvl 
| Public : conflvl.

Inductive intlvl : Set := 
  High : intlvl 
| Low : intlvl.

Inductive basic_type : Set :=
  IntT : basic_type 
| EncT : keyname -> prim_type -> basic_type

with prim_type : Set := 
  PrimT : basic_type -> conflvl -> intlvl -> prim_type.


Definition key_type_env : Set := env_skel keyname conflvl.
Definition var_type_env : Set := env_skel nat prim_type.


(* values *)

Inductive integer : Set :=
  IntV : nat -> integer.

Inductive key : Set := 
  KeyV : nat -> key.

Definition key_env := env_skel keyname key.
Definition var_env := env_skel nat integer.

Definition blind := stream_skel nat.
Definition blind_env := env_skel nat blind.

Axiom prim_enc : keyname -> integer -> integer -> integer.
Axiom prim_dec : keyname -> integer -> integer.


(* type rules *)
 
(* WF-INT *)

Inductive wf_value : integer -> prim_type -> Prop :=
  WF_INT : forall n : integer,
	    forall clvl : conflvl, 
             forall ilvl : intlvl, wf_value n (PrimT IntT clvl ilvl)

| WF_ENC : forall n : integer,
	   forall tau : prim_type,
	   forall k   : keyname,
           forall clvl : conflvl,
           forall ilvl : intlvl,
            forall n2 : integer, (* these two are extistential quantifiers since they don't occur *)
 	    forall n3 : integer, (*  on the right hand side of the implication *)
              n = prim_enc k n2 n3 /\ wf_value n3 tau ->  wf_value n (PrimT (EncT k tau) clvl ilvl).


Inductive wf_keyenv : key_env -> key_type_env -> Prop :=
  WF_EMPTY_KE : wf_keyenv (EmptyE keyname key) (EmptyE keyname conflvl)
| WF_KE       : forall ke  : key_env,
		forall kte : key_type_env, 
		forall kn  : keyname,
		forall k   : key,
		forall s   : conflvl,
		
		(* demand that kn is not in the domain of ke *)
		(forall k', not (env_lookup keyname key ke kn k')) ->
		(* demand that kn is not in the domain of kte *)
		(forall s', not (env_lookup keyname conflvl kte kn s')) ->
		wf_keyenv ke kte -> wf_keyenv (ConsE keyname key kn k ke) (ConsE keyname conflvl kn s kte).

Inductive wf_varenv : var_env -> var_type_env -> Prop :=
  WF_EMPTY_VE : wf_varenv (EmptyE nat integer) (EmptyE nat prim_type)
| WF_VE       : forall ve  : var_env,
		forall vte : var_type_env,
		forall x   : nat,
		forall n   : integer,
		forall t   : prim_type,

		(forall n', not (env_lookup nat integer ve x n')) ->
		(forall t', not (env_lookup nat prim_type vte x t')) ->
		wf_value n t ->
		wf_varenv ve vte ->
		wf_varenv (ConsE nat integer x n ve) (ConsE nat prim_type x t vte).

(* wf_nonce_env skipped *)


(* Inductive wf_env : *)

(* expressions *)


Inductive expr : Set := 
  SynConst : nat -> expr 
| SynVar   : nat -> expr
| SynOp    : expr -> expr -> expr
| SynEnc   : keyname -> expr -> expr
| SynDec   : keyname -> expr -> expr.

Inductive cmd : Set :=
  SynSkip   : cmd
| SynVarAss : nat -> expr -> cmd
| SynSeq    : cmd -> cmd -> cmd
| SynIf     : expr -> cmd -> cmd -> cmd
| SynWhile  : expr -> cmd -> cmd.
 




(* expression type rules *)

Inductive type_expr : expr -> key_type_env -> var_type_env -> prim_type -> Prop :=
  TYPE_INT : 
	forall n : nat,
	forall kte : key_type_env,
	forall vte : var_type_env,
	type_expr (SynConst n) kte vte (PrimT IntT Public High)

| TYPE_VAR:
	forall x : nat,
	forall kte : key_type_env,
	forall vte : var_type_env,	

	forall t : prim_type,

	env_lookup nat prim_type vte x t ->
	type_expr (SynVar x) kte vte t.

(* expression semantics *)

Inductive eval_expr : expr -> key_env -> var_env -> blind_env -> blind_env -> integer -> Prop :=
  EVAL_INT : 
	forall ke : key_env,
	forall ve : var_env,
	forall be : blind_env,
	forall n  : nat,
	eval_expr (SynConst n) ke ve be be (IntV n)

| EVAL_VAR :
	forall ke : key_env,
	forall ve : var_env,
	forall be : blind_env,
	forall x  : nat,
	forall n  : integer,
	
	env_lookup nat integer ve x n ->
		eval_expr (SynVar x) ke ve be be n. 	

Lemma lookup_empty :
	forall A B : Set,
	forall e : env_skel A B,
	forall a : A,
	forall b : B,

	env_lookup A B e a b -> EmptyE A B <> e.

Proof.
	intros.
	inversion H.
	  discriminate.
 	  discriminate.
Qed.	


Lemma pres_var :
	forall ve : var_env,
	forall vte : var_type_env,
	
	forall x : nat,
	forall n : integer,
	forall t : prim_type,

	wf_varenv ve vte -> env_lookup nat integer ve x n -> env_lookup nat prim_type vte x t -> wf_value n t.

Proof.

intros.
induction H.
 inversion H0.
 inversion H0.
  inversion H1.
   rewrite <- H14.
     rewrite <- H9.
     trivial.
   absurd (env_lookup nat prim_type vte x t).
    rewrite <- H5.
      apply H2.
    trivial.
  inversion H1.
   absurd (env_lookup nat integer ve x n).
    rewrite <- H11.
      apply H.
    trivial.
   apply IHwf_varenv.
    trivial.
    trivial.
Qed.



Theorem pres_expr : 
	forall e : expr,
	forall ke : key_env, 
	forall ve : var_env,
	forall kte : key_type_env,	
	forall vte : var_type_env,
	forall t   : prim_type,
	forall be  : blind_env,
	forall be' : blind_env,
	forall v   : integer,

	wf_keyenv ke kte ->
	wf_varenv ve vte ->

	type_expr e kte vte t ->
	eval_expr e ke  ve be be' v ->
	
	wf_value v t.

(* Time for a Lemma! 	wf_varenv ve vte -> lookup nat integer ve x n -> lookup nat primt_type vte x t ->
	wf_val x t *)


Proof.
(*	
	intros.
	induction H1.
		(* integer case *)
		apply WF_INT.
Qed.
		
*)