
-- # import prelude
-- # import nat
-- # import fin

-- The untyped language ---------------------------------------------------

infixl 20 @

data Term : Nat -> Set where
  var	(n : Nat) : Fin n -> Term n
  λ	(n : Nat) : Term (suc n) -> Term n
  (@)	(n : Nat) : Term n -> Term n -> Term n

data Typ : Nat -> Set where
  ★  (n : Nat) : Typ n
  Π  (n : Nat) : (A : Typ n) -> (B : Typ (suc n)) -> Typ n
  El (n : Nat) : Term n -> Typ n

-- Shifting
shiftFrom (n : Nat) : Fin (suc n) -> Term n -> Term (suc n)
shiftFrom i (var j)   = var (thin i j)
shiftFrom i (λ t)     = λ (shiftFrom (fsuc i) t)
shiftFrom i (s @ t)   = shiftFrom i s @ shiftFrom i t

shift (n : Nat) : Term n -> Term (suc n)
shift = shiftFrom fzero

shiftFromT (n : Nat) : Fin (suc n) -> Typ n -> Typ (suc n)
shiftFromT i ★	     = ★
shiftFromT i (Π A B) = Π (shiftFromT i A) (shiftFromT (fsuc i) B)
shiftFromT i (El t)  = El (shiftFrom i t)

shiftT (n : Nat) : Typ n -> Typ (suc n)
shiftT = shiftFromT fzero

-- Substitution
substitute (n : Nat) : Fin (suc n) -> Term n -> Term (suc n) -> Term n
substitute i s (var j) with thinView i j =>
  i s (var i) hit		= s
  i s (var (thin i k)) (miss k)	= var k
substitute i s (λ t)		= λ (substitute (fsuc i) (shift s) t)
substitute i s (t @ u)		= substitute i s t @ substitute i s u

substituteT (n : Nat) : Fin (suc n) -> Term n -> Typ (suc n) -> Typ n
substituteT i s ★	= ★
substituteT i s (Π A B)	= Π (substituteT i s A) (substituteT (fsuc i) (shift s) B)
substituteT i s (El t)	= El (substitute i s t)

app (n : Nat) : Term n -> Term n -> Term n
app (λ t) u	= substitute fzero u t
app (var i) u	= var i @ u
app (s @ t) u	= s @ t @ u

infixl 10 &

data Ctx : Nat -> Set where
  ε	      : Ctx 0
  (&) (n : Nat) : Ctx n -> Typ n -> Ctx (suc n)

-- Context lookup
(!) (n : Nat) : Ctx n -> Fin n -> Typ n
(Γ & A) ! fzero	  = shiftT A
(Γ & A) ! fsuc i  = shiftT (Γ ! i)

-- Some trickery
infix 2 ⊢
(⊢) (n : Nat)(R : Set) : Ctx n -> (Ctx n -> R) -> R
Γ ⊢ f = f Γ

-- The typing rules -------------------------------------------------------

mutual

  infix 4 ∈, ≡, ≐
  infix 3 ::

  type (n : Nat) : Typ n -> Ctx n -> Set
  type A = \Γ -> IsType _ Γ A

  (∈) (n : Nat) : Term n -> Typ n -> Ctx n -> Set
  t ∈ A = \Γ -> HasType _ Γ t A

  (≡) (n : Nat) : Typ n -> Typ n -> Ctx n -> Set
  A ≡ B = \Γ -> TypeEq _ Γ A B

  (≐) (n : Nat) : Term n -> Term n -> Typ n -> Ctx n -> Set
  s ≐ t = \A Γ -> TermEq _ Γ s t A

  (::) (n : Nat) : (Typ n -> Ctx n -> Set) -> Typ n -> Ctx n -> Set
  f :: A = f A

  context (n : Nat) : Ctx n -> Set
  context = \Γ -> WFCtx _ Γ

  -- Well formed context
  data WFCtx : (n : Nat) -> Ctx n -> Set where

    wfε	  : ---------
	    context ε


    wfExt (n : Nat)(Γ : Ctx n)(A : Typ n) :

	 context Γ  ->  Γ ⊢ type A
      -> -------------------------
	      context (Γ & A)

  -- Is type
  data IsType : (n : Nat) -> Ctx n -> Typ n -> Set where

    setIsType (n : Nat)(Γ : Ctx n) :
    
	      context Γ
	  -> ----------
	     Γ ⊢ type ★


    piIsType (n : Nat)(Γ : Ctx n)(A : Typ n)(B : Typ (suc n)) :

	 Γ ⊢ type A  ->  Γ & A ⊢ type B
      -> ------------------------------
	       Γ ⊢ type (Π A B)


    elIsType (n : Nat)(Γ : Ctx n)(t : Term n) :

	    Γ ⊢ t ∈ ★
      -> ---------------
	 Γ ⊢ type (El t)

  -- Has type
  data HasType : (n : Nat) -> Ctx n -> Term n -> Typ n -> Set where

    varType (n : Nat)(Γ : Ctx n)(i : Fin n) :

	       context Γ
	-> -----------------
	   Γ ⊢ var i ∈ Γ ! i


    λType (n : Nat)(Γ : Ctx n)(t : Term (suc n))(A : Typ n)(B : Typ (suc n)) :

	    Γ & A ⊢ t ∈ B
	-> ---------------
	   Γ ⊢ λ t ∈ Π A B


    appType (n : Nat)(Γ : Ctx n)(s, t : Term n)(A : Typ n)(B : Typ (suc n)) :

	     Γ ⊢ s ∈ Π A B  ->  Γ ⊢ t ∈ A
	-> ---------------------------------
	   Γ ⊢ s @ t ∈ substituteT fzero t B


    convType (n : Nat)(Γ : Ctx n)(t : Term n)(A,B : Typ n) :

	   Γ ⊢ t ∈ A  ->  Γ ⊢ A ≡ B
	-> ------------------------
		  Γ ⊢ t ∈ B

  -- Type equality
  data TypeEq : (n : Nat) -> Ctx n -> Typ n -> Typ n -> Set where

    ★Eq (n : Nat)(Γ : Ctx n) :

	   context Γ
	-> ---------
	   Γ ⊢ ★ ≡ ★

    ΠEq (n : Nat)(Γ : Ctx n)(A,A' : Typ n)(B,B' : Typ (suc n)) :

	   Γ ⊢ A ≡ A'  ->  Γ & A ⊢ B ≡ B'
	-> ------------------------------
	        Γ ⊢ Π A B ≡ Π A' B'

    ElEq (n : Nat)(Γ : Ctx n)(s,t : Term n) :

	    Γ ⊢ s ≐ t :: ★
	-> ---------------
	   Γ ⊢ El s ≡ El t

  -- Term equality (typed)
  -- Actually we need a spine view to do typed conversion!
  data TermEq : (n : Nat) -> Ctx n -> Term n -> Term n -> Typ n -> Set where

    varEq (n : Nat)(Γ : Ctx n)(i : Fin n) :

      --------------------------
      Γ ⊢ var i ≐ var i :: Γ ! i


    funEq (n : Nat)(Γ : Ctx n)(f,g : Term n)(A : Typ n)(B : Typ (suc n)) :

	 Γ & A ⊢ app (shift f) (var fzero) ≐ app (shift g) (var fzero) :: B
      -> ------------------------------------------------------------------
			      Γ ⊢ f ≐ g :: Π A B

