
module Plugins.FOL.Solve ( solveGoals ) where

import Control.Monad
import Control.Monad.Trans	( liftIO )
import Data.List		( isPrefixOf, isSuffixOf )
import System.Directory		( removeFile )
import System.Environment	( getEnv )
import Text.PrettyPrint

import IO.Pipes			( invokeWithTimeOut_, invoke_ )

import Interface.Options	( Options(..) )

import Syntax.Pretty
import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Options

import Utils.Error
import Utils.Monads
import Utils.Lists

import Plugins.FOL.Types
import Plugins.FOL.Errors
import Plugins.FOL.Options
import Plugins.FOL.Otter	( makeOtterFile, translateFormulas )
import Plugins.FOL.Gandalf

-- Solving ----------------------------------------------------------------

otterFileName :: FOLGoal -> FilePath
otterFileName (FOLGoal name _ _ _ _) = ".fol." ++ name ++ ".otter"

getProgram :: String -> String -> Maybe String -> IO String
getProgram env def opt =
    maybe   (getEnv env `catch` \ _ -> return def)
	    return
	    opt

solveGoals :: FOLOptions -> [FOLGoal] -> TC ()
solveGoals fol_opts goals =
    do  prover <- liftIO $ getProgram "GANDALF" "gandalf" (optProver fol_opts)
	finder <- liftIO $ getProgram "PARADOX" "paradox" (optFinder fol_opts)
	opts <- commandLineOptions
	when (optVerbose opts >= 2) $
	    liftIO $ putStrLn $ "Solving first order goals using " ++ prover
	gs <- concat <$> mapM (solve opts prover finder) (reverse goals)
	case gs of
	    []	-> return ()
	    _	-> failure $ failedToProveFirstOrderGoalsError gs
    where
	solve :: Options -> String -> String -> FOLGoal -> TC [(FOLGoal, Maybe CounterModel)]
	solve opts prover finder g =
	    do  r <- liftIO $ solveFOLGoal
				    (optTimeout fol_opts)
				    prover
				    finder
				    (optCounterModels fol_opts)
				    (optModelSize fol_opts)
				    (optVerbose opts)
				    (optKeepFiles fol_opts)
				    g
		return $ case r of
		    Right ()	-> []
		    Left m	-> [(g, m)]

{--------------------------------------------------------------------------
    Calling gandalf and paradox
 --------------------------------------------------------------------------}

solveFOLGoal :: Int	->  -- Time out
		String	->  -- gandalf
		String	->  -- paradox
		Bool	->  -- call paradox on failure?
		Int	->  -- maximum model size
		Int	->  -- verbosity
		Bool	->  -- keep files?
		FOLGoal ->  -- goal
		IO (Either (Maybe CounterModel) ())
		    -- Right () = ok
		    -- Left m   = fail, counter model: m
solveFOLGoal timeout prover finder models model_size verbose keepFiles
	    g@(FOLGoal _ (lams,pre) goal vs v) =

    do	-- Clausification
	let goal' : pre' = translateFormulas (Not goal : lams ++ pre)

	-- Create the input file
	let filename  = otterFileName g
	    otterFile = makeOtterFile (concat pre', goal')
	writeFile filename otterFile

	printGoal verbose g
	printInput verbose otterFile

	-- Call gandalf
	(output,err) <- invokeWithTimeOut_ timeout prover [filename]

	verifyGandalfErrors err

	printOutput verbose output err

	-- Check for success
	let ok = containsProof output

	printStatus verbose ok output

	-- Look for a counter model
	let findModel
		| ok	    = return $ Right ()
		| models    = Left <$> findCounterModel	verbose finder
							model_size filename
		| otherwise = return $ Left Nothing
	r <- findModel

	-- Remove file
	unless keepFiles $ removeFile filename

	return r
    where
	containsProof = elem "***GANDALF_FOUND_A_REFUTATION***" . lines

	verifyGandalfErrors err =
	    case lines err of
		s:_ | "forkProcess" `isSubstringOf` s
			-> ioError $ userError $
					"Failed to call first order prover '"
					++ prover ++ "'. Probable cause: prover not found."
		_	-> return ()


{--------------------------------------------------------------------------
    Finding a counter model
 --------------------------------------------------------------------------}

-- | Call paradox to find a counter model to a FOL goal
findCounterModel :: Int -> String -> Int -> FilePath -> IO (Maybe CounterModel)
findCounterModel verbose finder model_size filename =

    do	printCounterModelSearchStarted verbose

	-- Call paradox
	(output,err) <- invoke_ finder
			    [ "--print", "Model"
			    , "--sizes", "1.." ++ show model_size
			    , filename
			    ]

	printParadoxOutput verbose output err

	-- Parse paradox output
	let m = extractModel output

	printCounterModel verbose m

	return m

-- | Extract the model from paradox's output
extractModel :: String -> Maybe CounterModel
extractModel s
    | null m	= Nothing
    | otherwise	= Just m
    where
	modelStart  = isPrefixOf "== Model"
	anyStart    = isPrefixOf "=="
	
	m = filter (not . null)
	    $ takeWhile (not . anyStart)
	    $ drop 1
	    $ dropWhile (not . modelStart)
	    $ lines s

{--------------------------------------------------------------------------
    Debug output
 --------------------------------------------------------------------------}

printGoal verbose (FOLGoal name (lams,pre) goal vs v)
    | verbose == 2  =
	print $ vcat $
		    [ text ""
		    , text "--- Trying to prove goal " <+> text name <> text ":"
		    ] ++
		    map (nest 6)
			(	map pretty vs
			++	[ text $ replicate 40 '-'
			    , pretty v 
			    ]
			)
    | verbose >= 3  =
	print $ vcat $
		    [ text ""
		    , text "--- Trying to prove goal" <+> text name <> text ":"
		    ] ++
		    let pr (v,p) = pretty v <> text ":" $+$ 
				   (nest 2 $ pretty p)
			plam = case lams of
				    []  -> []
				    _   -> [ text "Lifted lambdas:" $+$
						nest 2 (vcat $ map pretty lams)
					    ]
		    in map (nest 6)
			(	plam
			++	map pr (zip vs pre)
			++	[ text $ replicate 40 '-'
			    , pr (v, goal)
			    ]
			)
    | otherwise	    = return ()

printInput verbose otterFile =
    when (verbose >= 5) $
	putStrLn $ "\n--- input\n" ++ otterFile

printOutput verbose output err =
    do	when (verbose >= 5) $
	    do	putStrLn $ "--- output"
		putStr output
	when (verbose >= 4) $
	    do	putStrLn $ "--- output on stderr"
		putStr err

printStatus verbose ok output =
    when (verbose >= 2) $
	if ok
	    then do putStrLn "+++ Proof succeeded:"
		    when (verbose >= 3) $
			putStr
			    $ unlines
			    $ map ("    "++)
			    $ lines
			    $ gandalfProof output
	    else putStrLn "--- Proof failed"

printCounterModelSearchStarted verbose =
    when (verbose >= 2) $ putStrLn "--- Looking for counter model"

printParadoxOutput verbose output err =
    do	when (verbose >= 4) $
	    do  putStrLn "--- paradox stdout"
		putStrLn output
	when (verbose >= 2 && not (null err)) $
	    do  putStrLn "--- paradox stderr"
		putStrLn err

printCounterModel verbose m =
    when (verbose >= 2) $
	case m of
	    Nothing	-> putStrLn "  No counter model found"
	    Just ss	->
		do  putStrLn "  Found counter model:"
		    mapM_ (putStrLn . ("    "++)) ss
