{-# OPTIONS -fglasgow-exts #-}

{- |
    The default mechanism for translating application into FOL is to use the
    function symbol _app. However if a function is always used with the same
    arity, it is safe to translate is as a function symbol of that arity. This
    is what this module does.
-}

module Plugins.FOL.Arities where

import Utils.Pairs
import Utils.Map hiding (map)

import Plugins.FOL.Types
import Plugins.FOL.Primitive

{--------------------------------------------------------------------------
    Optimize function applications
 --------------------------------------------------------------------------}

removeBinaryApps :: (ComputeArities a, Optimize a) => a -> a
removeBinaryApps x = optimize (computeArities x) x

{--------------------------------------------------------------------------
    FOL term app view
 --------------------------------------------------------------------------}

data FOLAppTerm = FOLAppFun String [FOLTerm]
		| FOLAppVar String [FOLTerm]

folAppView :: FOLTerm -> FOLAppTerm
folAppView (FOLFun app [s,t])
    | app == kwApp  = folApp (folAppView s) t
folAppView (FOLFun f ts)    = FOLAppFun f ts
folAppView (FOLVar x)	    = FOLAppVar x []

folApp (FOLAppFun f ts) t = FOLAppFun f (ts ++ [t])
folApp (FOLAppVar x ts) t = FOLAppVar x (ts ++ [t])

{--------------------------------------------------------------------------
    Arities
 --------------------------------------------------------------------------}

type Arities = Map String (Maybe Int)

combine (Just n) (Just m)
    | n == m	= Just n
combine _ _	= Nothing

unit x ts = singleton x (Just $ length ts)
plus = unionWith combine
pluss = foldr plus empty

arity :: Arities -> String -> Maybe Int
arity ar x = findWithDefault Nothing x ar

{--------------------------------------------------------------------------
    Computing the arities
 --------------------------------------------------------------------------}

class ComputeArities a where
    computeArities :: a -> Arities

instance ComputeArities a => ComputeArities [a] where
    computeArities = pluss . map computeArities

instance (ComputeArities a, ComputeArities b) =>
	    ComputeArities (a,b) where
    computeArities (x,y) = pluss    [ computeArities x
				    , computeArities y
				    ]

instance (ComputeArities a, ComputeArities b, ComputeArities c) =>
	    ComputeArities (a,b,c) where
    computeArities (x,y,z) = pluss  [ computeArities x
				    , computeArities y
				    , computeArities z
				    ]

instance ComputeArities Arities where
    computeArities = id

instance ComputeArities Formula where
    computeArities p =
	case p of
	    Predicate _ ts  -> computeArities ts
	    And p q	    -> computeArities (p,q)
	    Or p q	    -> computeArities (p,q)
	    Not p	    -> computeArities p
	    Implies p q	    -> computeArities (p,q)
	    Equiv p q	    -> computeArities (p,q)
	    ForAll _ f	    -> computeArities (f $ FOLVar "_")
	    Exists _ f	    -> computeArities (f $ FOLVar "_")
	    TRUE	    -> empty
	    FALSE	    -> empty

instance ComputeArities FOLTerm where
    computeArities t =
	case folAppView t of
	    FOLAppFun f ts  -> computeArities (unit f ts, ts)
	    FOLAppVar _ ts  -> computeArities ts

{--------------------------------------------------------------------------
    Updating the formulae
 --------------------------------------------------------------------------}

class Optimize a where
    optimize :: Arities -> a -> a

instance Optimize a => Optimize [a] where
    optimize ar = map (optimize ar)

instance (Optimize a, Optimize b) => Optimize (a,b) where
    optimize ar = optimize ar -*- optimize ar

instance (Optimize a, Optimize b, Optimize c) => Optimize (a,b,c) where
    optimize ar (x,y,z) = (optimize ar x, optimize ar y, optimize ar z)

instance Optimize Formula where
    optimize ar p =
	case p of
	    Predicate x ts  -> Predicate x $ optimize ar ts
	    And p q	    -> uncurry And $ optimize ar (p,q)
	    Or p q	    -> uncurry Or $ optimize ar (p,q)
	    Not p	    -> Not $ optimize ar p
	    Implies p q	    -> uncurry Implies $ optimize ar (p,q)
	    Equiv p q	    -> uncurry Equiv $ optimize ar (p,q)
	    ForAll x f	    -> ForAll x $ \t -> optimize ar (f t)
	    Exists x f	    -> Exists x $ \t -> optimize ar (f t)
	    TRUE	    -> TRUE
	    FALSE	    -> FALSE

instance Optimize FOLTerm where
    optimize ar t =
	case folAppView t of
	    FOLAppFun f ts  ->
		case arity ar f of
		    Just _  -> FOLFun f $ optimize ar ts
		    _	    -> apps (FOLFun f []) $ optimize ar ts
	    FOLAppVar x ts  -> apps (FOLVar x) $ optimize ar ts
	where
	    apps = foldl (\x y -> app [x,y])

