
module Plugins.FOL.Solve ( solveGoals ) where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad
import Control.Monad.Trans	( liftIO )
import Data.List
import Data.Char
import System.Directory
import System.Environment	( getEnv )
import Text.PrettyPrint

import IO.Invoke

import Interface.Options	( Options(..) )
import Interface.ChaseImports	( findModule )

import Syntax.Pretty
import Syntax.Position
import Syntax.Internal
import Syntax.Values

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Options
import TypeChecker.Definitions ( allDefinitions )  -- for slicing
import TypeChecker.UnfoldConstants

import Utils.Error
import Utils.Monads
import Utils.Lists
import Utils.Files

import Plugins.FOL.Types
import Plugins.FOL.Pretty
import Plugins.FOL.Errors
import Plugins.FOL.Options
import Plugins.FOL.Otter	( makeOtterFile, translateFormulas )
import Plugins.FOL.Gandalf
import Plugins.FOL.Santa
import Plugins.FOL.Arguments

-- Solving ----------------------------------------------------------------

-- Creating the first order file
makeFOLFile :: Way -> String -> Range -> Arguments -> [FilePath] -> Int ->
	       Bool -> Value -> Formula -> [Formula] -> TC (FilePath, String)
makeFOLFile Otter _ r _ _ _ verbose _ goal pre =
    do	-- Clausification
	let goal' : pre' = translateFormulas
			    (Not goal : pre)

	-- Create the input file
	let filename  = otterFileName r
	    otterFile = makeOtterFile (concat pre', goal')
	liftIO $ writeFile filename otterFile
	return (filename, otterFile)

makeFOLFile way santa r args dirs verbose keep_files v goal pre =
    do	inc <- mkCoverIncludes r args dirs verbose keep_files v
	let sfilename	= santaFileName r
	    santaFile	= makeSantaFile inc pre goal
	liftIO $ writeFile sfilename santaFile
	(out,err) <- liftIO $ invoke_ santa [sfilename, outflag, filename]
	printSantaOutput verbose out err
	force out   -- make sure thing actually happen
	input <- liftIO $ readFile filename
	unless keep_files $ liftIO $ mapM_ removeFile $ sfilename : inc
	return (filename, input)
    where
	(filename, outflag) =
	    case way of
		SantaOtter  -> (otterFileName r, "--otter-clauses")
		SantaTSTP   -> (tstpFileName r, "--tstp-clauses")
		Otter	    -> error "the otter backend requires Santa"

-- | TODO: a bit ad hoc at the moment
mkCoverIncludes :: Range -> Arguments -> [FilePath] -> Int -> Bool -> Value -> TC [FilePath]
mkCoverIncludes r args dirs verbose keep_files v
    | null (includes args)  = return []
    | otherwise		    =
	do  corefile:corefiles <- mapM createCore $ includes args
	    sliceMe <-
		filter (isHaskellName . nameString) <$>
		    (allDefinitions =<< unfoldAllButSomeConstants (freeze args) v)
	    let dontSliceMe = freeze args
	    let tstpfile = mkFileName r ".cover"
	    let cmdargs	= 
		    ["--bottom", "-d","-o",tstpfile] ++
		    (concatMap (\d -> ["-s",unquote $ nameString d]) sliceMe) ++
		    (concatMap (\f -> ["-i",f]) corefiles) ++
		    [corefile]
	    when (verbose >= 4) $
		liftIO $ putStrLn $ "> CoverTranslator " ++ unwords cmdargs
	    (out,err) <- liftIO $ invoke_ "CoverTranslator" cmdargs
	    force out
	    when (verbose >= 4) $ liftIO $
		do  putStrLn "--- CoverTranslator out"
		    putStr out
		    putStrLn "--- CoverTranslator err"
		    putStr err
		    when (verbose >= 5) $
			do  putStrLn "--- CoverTranslator result"
			    putStr =<< readFile tstpfile
	    return [tstpfile]
    where
	unquote ('"':x) = "'" ++ init x ++ "'"
	unquote x	= x

	-- just an approximation
	isHaskellName ('"':c:x) = isUpper c && elem '.' x
	isHaskellName _ = False

	createCore inc =
	    liftIO $
	    do	mHsFile	    <- findModule dirs inc [".hs"]
		mHcrFile    <- findModule dirs inc [".hcr"]
		case mHcrFile of
		    Nothing ->
			case mHsFile of
			    Nothing	-> fail $ "Haskell module '" ++ inc ++ "' not found in " ++ concat (intersperse ", " dirs)
			    Just hsfile	->
				do  (out,err) <- invoke_ "ghc" ["-c","-fext-core",hsfile]
				    force out
				    when (verbose >= 4) $
					do  putStrLn "--- ghc out"
					    putStr out
					    putStrLn "--- ghc err"
					    putStr err
				    return $ setExtension ".hcr" hsfile
		    Just file	-> return file

mkFileName :: Range -> String -> FilePath
mkFileName r ext = ".fol." ++ clean (show r) ++ ext
    where
	clean = concatMap f
	    where
		f '/' = "-"
		f ':' = "="
		f c = [c]

otterFileName	= flip mkFileName ".otter"
santaFileName	= flip mkFileName ".santa"
tstpFileName	= flip mkFileName ".tstp"

getProgram :: String -> String -> Maybe String -> IO String
getProgram env def opt =
    maybe   (getEnv env `catch` \ _ -> return def)
	    return
	    opt

solveGoals :: FOLOptions -> [FOLGoal] -> TC ()
solveGoals fol_opts goals =
    do  prover <- liftIO $ getProgram "GANDALF" "gandalf" (optProver fol_opts)
	finder <- liftIO $ getProgram "PARADOX" "paradox" (optFinder fol_opts)
	santa  <- liftIO $ getProgram "SANTA" "santa" (optSanta fol_opts)
	opts <- commandLineOptions
	when (optVerbose opts >= 2) $
	    liftIO $ putStrLn $ "Solving first order goals using " ++ prover
	gs <- concat <$> mapM (solve opts prover finder santa) (reverse goals)
	case gs of
	    []	-> return ()
	    _	-> failure $ failedToProveFirstOrderGoalsError gs
    where
	solve :: Options -> String -> String -> String -> FOLGoal -> TC [(FOLGoal, Maybe CounterModel)]
	solve opts prover finder santa g =
	    do  r <- solveFOLGoal
				    (optTimeout fol_opts)
				    prover
				    finder
				    (optCounterModels fol_opts)
				    (optModelSize fol_opts)
				    santa
				    (optWay fol_opts)
				    (optVerbose opts)
				    (optKeepFiles fol_opts)
				    (optIncludeDirs opts)
				    g
		return $ case r of
		    Right ()	-> []
		    Left m	-> [(g, m)]

{--------------------------------------------------------------------------
    Calling gandalf and paradox
 --------------------------------------------------------------------------}

solveFOLGoal :: Int	    ->  -- Timeout
		String	    ->  -- gandalf
		String	    ->  -- paradox
		Bool	    ->  -- call paradox on failure?
		Int	    ->  -- maximum model size
		String	    ->  -- santa
		Way	    ->  -- how and what to produce
		Int	    ->  -- verbosity
		Bool	    ->  -- keep files?
		[FilePath]  ->	-- include dirs
		FOLGoal	    ->  -- goal
		TC (Either (Maybe CounterModel) ())
		    -- Right () = ok
		    -- Left m   = fail, counter model: m
solveFOLGoal timeout prover finder models model_size
	     santa way verbose keepFiles dirs
	     g@(FOLGoal r args (lams,pre) goal vs v) =

    do	(filename, input) <- makeFOLFile way santa r args dirs verbose keepFiles
					 v goal (lams ++ pre)

	printGoal verbose g
	printInput verbose input

	-- Call gandalf
	(output,err) <- liftIO $ invoke_ prover ["-time", show $ div timeout 1000 `max` 1, filename]

	verifyGandalfErrors err

	printOutput verbose output err

	-- Check for success
	let ok = containsProof output

	printStatus verbose way ok output

	-- Look for a counter model
	let findModel
		| ok	    = return $ Right ()
		| models    = Left <$> findCounterModel	verbose finder
							model_size filename
		| otherwise = return $ Left Nothing
	r <- findModel

	-- Remove file
	unless keepFiles $ liftIO $ removeFile filename

	return r
    where
	containsProof = elem "***GANDALF_FOUND_A_REFUTATION***" . lines

	verifyGandalfErrors err =
	    liftIO $
	    case lines err of
		s:_ | "forkProcess" `isSubstringOf` s
			-> ioError $ userError $
					"\nFailed to call first order prover '"
					++ prover ++ "'.\nProbable cause: prover not found.\n" ++
					err
		_	-> return ()


{--------------------------------------------------------------------------
    Finding a counter model
 --------------------------------------------------------------------------}

-- | Call paradox to find a counter model to a FOL goal
findCounterModel :: Int -> String -> Int -> FilePath -> TC (Maybe CounterModel)
findCounterModel verbose finder model_size filename =

    do	printCounterModelSearchStarted verbose

	-- Call paradox
	(output,err) <- liftIO $ invoke_ finder
			    [ "--print", "Model"
			    , "--sizes", "1.." ++ show model_size
			    , filename
			    ]

	printParadoxOutput verbose output err

	-- Parse paradox output
	let m = extractModel output

	printCounterModel verbose m

	return m

-- | Extract the model from paradox's output
extractModel :: String -> Maybe CounterModel
extractModel s
    | null m	= Nothing
    | otherwise	= Just m
    where
	modelStart  = isPrefixOf "== Model"
	anyStart    = isPrefixOf "=="
	
	m = filter (not . null)
	    $ takeWhile (not . anyStart)
	    $ drop 1
	    $ dropWhile (not . modelStart)
	    $ lines s

{--------------------------------------------------------------------------
    Debug output
 --------------------------------------------------------------------------}

printGoal verbose (FOLGoal r args (lams,pre) goal vs v)
    | verbose == 2  =
	liftIO $ print $ vcat $
		    [ text ""
		    , text "--- Trying to prove goal at" <+> pretty r <> text ":"
		    ] ++
		    map (nest 6)
			(	map pretty vs
			++	[ text $ replicate 40 '-'
			    , pretty v 
			    ]
			)
    | verbose >= 3  =
	liftIO $ print $ vcat $
		    [ text ""
		    , text "--- Trying to prove goal at" <+> pretty r <> text ":"
		    ] ++
		    let pr (v,p) = pretty v <> text ":" $+$ 
				   (nest 2 $ pretty p)
			plam = case lams of
				    []  -> []
				    _   -> [ text "Lifted lambdas:" $+$
						nest 2 (vcat $ map pretty lams)
					    ]
		    in map (nest 6)
			(	plam
			++	map pr (zip vs pre)
			++	[ text $ replicate 40 '-'
			    , pr (v, goal)
			    ]
			)
    | otherwise	    = return ()

printInput verbose otterFile =
    liftIO $
    when (verbose >= 6) $
	putStrLn $ "\n--- input\n" ++ otterFile

printOutput verbose output err =
    liftIO $
    do	when (verbose >= 5) $
	    do	putStrLn $ "--- output"
		putStr output
	when (verbose >= 4) $
	    do	putStrLn $ "--- output on stderr"
		putStr err

printStatus verbose way ok output =
    liftIO $
    when (verbose >= 2) $
	if ok
	    then do putStrLn "+++ Proof succeeded:"
		    let mangling = case way of
					Otter	-> AgdaLightMangling
					_	-> SantaMangling
		    when (verbose >= 3) $
			putStr
			    $ unlines
			    $ map ("    "++)
			    $ lines
			    $ gandalfProof mangling output
	    else putStrLn "--- Proof failed"

printCounterModelSearchStarted verbose =
    liftIO $
    when (verbose >= 2) $ putStrLn "--- Looking for counter model"

printSantaOutput verbose out err =
    liftIO $
    do	when (verbose >= 4) $
	    do  putStrLn "--- santa stdout"
		putStrLn out
	when (verbose >= 4 && not (null err)) $
	    do  putStrLn "--- santa stderr"
		putStrLn err

printParadoxOutput verbose output err =
    liftIO $
    do	when (verbose >= 4) $
	    do  putStrLn "--- paradox stdout"
		putStrLn output
	when (verbose >= 2 && not (null err)) $
	    do  putStrLn "--- paradox stderr"
		putStrLn err

printCounterModel verbose m =
    liftIO $
    when (verbose >= 2) $
	case m of
	    Nothing	-> putStrLn "  No counter model found"
	    Just ss	->
		do  putStrLn "  Found counter model:"
		    mapM_ (putStrLn . ("    "++)) ss
