{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances #-}

module Plugins.QuickCheck where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output
import Data.Map as Map hiding (Map)
import Data.List as List

import Control.Monad
import Control.Monad.Trans
import Control.Monad.Reader
import Control.Monad.State
import qualified Test.QuickCheck as Q

import Text.PrettyPrint
import System.Random
import System.IO hiding	    ( print, putStr, putStrLn )

import Syntax.Internal
import Syntax.Values
import Syntax.Pretty
import Syntax.Scope

import TypeChecker
import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Definitions
import TypeChecker.UnfoldConstants
import TypeChecker.Values
import TypeChecker.Haskell
import TypeChecker.Abstractions
import TypeChecker.Errors
import TypeChecker.Context
import TypeChecker.Monad.Context

import Compilers.Run

import Utils.Error
import Utils.Names
import Utils.Monads
import Utils.Pairs
import Utils.Map

import Plugins

-- The plugin -------------------------------------------------------------

plugin :: SomePlugin
plugin = mkPlugin $
    Plugin  { plName		= "qc"
            , plHelp            = helpQuickCheck
	    , plInit		= initQC
	    , plFinal		= testGoals
	    , plCheck		= checkQC
	    , plInfer		= \_ _ _ _ -> failure noInferenceForQCError
	    , plOptions		= simpleOptions []
	    , plDefaultOptions	= ()
	    }

-- Constants --------------------------------------------------------------

-- Generators
qc_Gen		= "Gen"
qc_return	= "returnGen"
qc_bind		= "bindGen"
qc_oneof	= "oneof"
qc_sized	= "sized"
qc_frequency	= "frequency"
qc_variant	= "variant"
qc_promote	= "promote"
qc_elements	= "elements"

-- Properties
qc_Tested	= "Tested"
qc_Prop		= "Prop"
qc_Eq		= "=="
qc_True		= "True"
qc_False	= "False"
qc_collect	= "collect"
qc_label	= "label"
qc_classify	= "classify"
qc_trivial	= "trivial"
qc_forAll	= "forAll"
qc_implies	= "==>"

-- Properties -------------------------------------------------------------

data QCProp = QCProp [(Name, Value, Type)] Value

-- Init -------------------------------------------------------------------

initQC _ = return []

-- Final ------------------------------------------------------------------

testGoals _ []	  = return ()
testGoals _ props =
    do	defs <- compileDefs Map.empty
	liftIO $ do putStrLn "=== Testing QuickCheck properties ===================="
		    hSetBuffering stdout NoBuffering
	mapM_ (testProp defs) props
    where
	testProp defs (QCProp gens v) =
	    do	liftIO $ print $ sep
			    [ text "Testing property"
			    , nest 2 $ sep [ text "\x2200"
					   , nest 2 $ fsep (punctuate comma $ List.map pr gens)
							<+> text "."
					   , nest 2 $ pretty $ instantiate v [ vId x | x <- reverse xs ]
					   ]
			    ]
		scope <- getScopeEnv
		let fa = getName scope qc_forAll
		prop  <- compileValue Map.empty $ mkProp fa gens v
		liftIO $ Q.quickCheck $ toProp scope $ evaluate defs prop
	    where
		toProp :: ScopeEnv -> NF -> Q.Property
		toProp scope = toHaskell (HEnv defs scope)
		xs = [ x | (x,_,_) <- gens ]

		pr (x,g,_) = prettyName x <+> text "\x2208" <+> pretty g

		mkProp _  []	       v = v
		mkProp fa ((x,g,w):gs) v =
		    vDef fa 1 `vApps` [unEl w, g, vLam $ mkAbs (nameString x) $ mkProp fa gs v]
		    

-- Checking ---------------------------------------------------------------

getTest _Tested t =
    do	v <- unfoldConstants t
	case vAppView v of
	    VAVDef c _ [p]
		| c == _Tested	-> return p
	    _			-> failure $ notAPropError t

checkQC :: a -> [QCProp] -> b -> [Term] -> Type -> TC (Value, [QCProp])
checkQC _ s _ ts t =
    do  scope     <- getScopeEnv
	let gen	    = getName scope qc_Gen
	    _Tested = getName scope qc_Tested
	(gens,g') <- mkGens _Tested (vDef gen 0) ts (unEl t)
	return (vTop, s ++ [QCProp gens g'])
    where
	mkGens _Tested d [] p =
	    do	g <- getTest _Tested p
		return ([],g)
	mkGens _Tested d (t:ts) p =
	    do	p' <- unfoldConstants p
		case vView p' of
		    VPi w f ->				-- TODO: cheating
			do  g	     <- checkType t (el (getSort w) $ d `vApply` unEl w)
			    x	     <- fresh (absName f)
			    (gs,p'') <- extendContext x (Abstract w)
					    $ mkGens _Tested d ts (unEl $ absBody f)
			    return ((x,g,w):gs, p'')
		    _	    -> failure $ Plugins.QuickCheck.tooManyArgumentsError p

instance Extractable Q.Property where
    toHaskell henv v =
	case v of
	    NCon c [_,v,w]
		| c =?= qc_Eq	-> Q.property $ v == w
	    NCon c [_,g,f]
		| c =?= qc_forAll   -> Q.forAll (toGen g) (toHaskell henv f :: NF -> Q.Property)
	    NCon c []
		| c =?= qc_True	    -> Q.property True
		| c =?= qc_False    -> Q.property False
	    NCon c [_,v,p]
		| c =?= qc_collect  -> Q.collect v (toProp p)
	    NCon c [_,l,p]
		| c =?= qc_label    -> Q.label (show l) (toProp p)
	    NCon c [_,b,l,p]
		| c =?= qc_classify -> Q.classify (toHaskell henv b) (show l)
						  (toProp p)
	    NCon c [b,p]
		| c =?= qc_trivial  -> Q.trivial (toHaskell henv b) (toProp p)
	    NCon c [p,q]
		| c =?= qc_implies  -> mkBoolProp p Q.==> toProp q
				    -- Note: I wonder if this is the right
				    -- choice? Perhaps it is, if we arrive
				    -- at something unrecognizable it is a
				    -- false property. At least it makes a
				    -- few things go through that would
				    -- otherwise yield an error (e.g.
				    -- RemoveInv in the Queue example).
	    _		    -> Q.property False --error $ "Not a quickCheckable property: " ++ show goal
	where
	    x =?= s = isName henv s x

	    app = nfApply (hDefinitions henv)

	    toProp :: NF -> Q.Property
	    toProp = toHaskell henv

	    toGen :: NF -> Q.Gen NF
	    toGen = toHaskell henv

	    mkBoolProp v =
		case v of
		    NCon c []
			| c =?= qc_True  -> True
			| c =?= qc_False -> False
		    NCon c [_, v, w]
			| c =?= qc_Eq	-> v == w
		    _			-> error $ "Not a decidable proposition: " ++ show v

instance Extractable (Q.Gen NF) where
--instance Extractable (Q.Gen NF) where
    toHaskell henv v =
	case v of
	    NCon c [_,w]
		| c =?= qc_return   -> return (toH w)
	    NCon c [_,_,g,k]
		| c =?= qc_bind	    -> do v <- toH g :: Q.Gen NF
					  toH k v
	    NCon c [_,w]
		| c =?= qc_elements  -> Q.elements (toH w)
		| c =?= qc_oneof     -> Q.oneof (toH w)
		| c =?= qc_frequency -> Q.frequency (toH w)
	    NCon c [_,f]
		| c =?= qc_sized     -> Q.sized (toH f)
	    NCon c [_, n, g]
		| c =?= qc_variant   -> Q.variant (toH n) (toH g)
	    NCon c [_,_,f]
		| c =?= qc_promote   -> fromHaskell henv <$> Q.promote (toH f :: NF -> Q.Gen NF)
	    _	-> error $ "Not a canonical generator: " ++ show v
	where
	    x =?= s = isName henv s x
	    toH x = toHaskell henv x

-- Errors -----------------------------------------------------------------

noInferenceForQCError :: ErrorFun
noInferenceForQCError e =
    e	{ errWhat   = fsep $ pwords "type inference failed"
	, errWhy    = fsep $ pwords "the type of a testable goal cannot be inferred"
	}

tooManyArgumentsError :: Value -> ErrorFun
tooManyArgumentsError v e =
    e	{ errWhat   = fsep $ pwords "QuickCheck was called with too many arguments"
	, errWhy    = fsep $ pwords "the goal" ++ [pretty v] ++
			     pwords "is not a function type"
	}

notAPropError :: Value -> ErrorFun
notAPropError v e =
    e	{ errWhat   = fsep $ pwords "Cannot quickCheck goal"
	, errWhy    = fsep $ pwords "the goal" ++ [pretty v] ++
			     pwords "is not on the form Tested P"
	}

-- Help -------------------------------------------------------------------

helpQuickCheck = "The QuickCheck plugin.\n"
