{
{-# OPTIONS -fglasgow-exts #-}
module Syntax.Lexer (lexer, Token(..), tex, code) where

import Data.List
import Data.Char

import Syntax.ParseMonad
import Syntax.Position

import Utils.Debug
import Utils.Unicode (isUnicodeId, isUnicodeOp)

}

$digit	    = 0-9
$idstart    = [ A-Z a-z ]
$symstart   = [ \- \! \# \$ \% \& \* \+ \/ \< \= \> \@ \^ \| \~ \? ]
$specialsym = [ \\ \: \. ]
$idchar	    = [ $symstart $idstart $digit ' _ ]
$namechar   = [ $idchar $specialsym ]
$endcomment = ~ $namechar

$white_nonl = $white # \n

@number	    = $digit+
@ident	    = $idstart $idchar* | _ $idchar+
-- operators starting with a special char cannot have an $idstart as second
-- char (to allow \x -> x and x:A)
@specialop  = $specialsym $namechar # [ $idstart $digit _ ] $namechar*
@operator   = $symstart $namechar* | @specialop
@plugin	    = @ident \- plugin

tokens :-

<tex>	 ^ \\ "begin{code}" \n	{ begin_ code }
<tex>	 ^ .* \n		{ withRange $ \_ -> TkTeX }
<tex>	 ^ .+			{ withRange $ \_ -> TkTeX }
<code>   ^ \\ "end{code}" \n	{ end_ }

<0,code,bol,layout,layout_left>
    $white_nonl+	;

<0,code,bol,layout,layout_left>
    "{-"		{ nestedComment }

-- Dashes followed by an operator symbol should be parsed as an operator.
<0,code,bol,layout,layout_left>
    "--"\-* $endcomment .*   ;
<0,code,bol,layout,layout_left>
    "--"\-* $		     ;
<0,code> \n		{ begin bol }
<bol>
    {
	\n		;
	()		{ beginningOfLine }
    }
<layout>
    {	-- \{ / { notFollowedBy '-' }  { popAnd openBrace }
	\n			    ;
	()			    { newLayoutContext }
    }
<layout_left> ()		{ layoutLeft }

<0,code> ISet			{ withRange_ TkISet }
<0,code> Set			{ withRange_ TkSet }
<0,code> Prop			{ withRange_ TkProp }
<0,code> Type			{ withRange_ TkType }
<0,code> data			{ withRange_ TkData }
<0,code> mutual			{ withLayout $ withRange_ TkMutual }
<0,code> abstract		{ withLayout $ withRange_ TkAbstract }
<0,code> class			{ withRange_ TkClass }
<0,code> instance		{ withRange_ TkInstance }
<0,code> extends		{ withRange_ TkExtends }
<0,code> rewrite		{ withLayout $ withRange_ TkRewrite }
<0,code> where			{ withLayout $ withRange_ TkWhere }
<0,code> with			{ withRange_ TkWith }
<0,code> let			{ withLayout $ withRange_ TkLet }
<0,code> do			{ withLayout $ withRange_ TkDo }
<0,code> sig			{ withRange_ TkSig }
<0,code> struct			{ withLayout $ withRange_ TkStruct }
<0,code> in			{ withRange_ TkIn }
<0,code> infix			{ withRange_ TkInfix }
<0,code> infixl			{ withRange_ TkInfixL }
<0,code> infixr			{ withRange_ TkInfixR }
<0,code> @plugin		{ withRange $ \r s -> TkPlugin (r, takeWhile ('-'/=) s) }
<0,code> @ident			{ withRange (curry TkId) }
<0,code> @number		{ withRange $ \r s -> TkInt (r,read s) }
<0,code> "_"			{ withRange_ TkUnderscore }    
<0,code> "#" $digit+		{ withRange $ \r s -> TkUniverse (r, read (tail s)) }
<0,code> "("			{ withRange_ TkLParen }
<0,code> ")"			{ withRange_ TkRParen }
<0,code> "\"			{ withRange_ TkLambda }
<0,code> "->"			{ withRange_ TkArrow }
<0,code> "<-"			{ withRange_ TkFrom }
<0,code> ","			{ withRange_ TkComma }
<0,code> ";"			{ withRange_ TkSemi }
<0,code> ":"			{ withRange_ TkColon }
<0,code> "=>"			{ withLayout $ withRange_ TkDblArrow }
<0,code> "="			{ withRange_ TkEqual }
<0,code> "."			{ withRange_ TkDot }
<0,code> "`"			{ withRange_ TkBackQuote }
<0,code> "{"			{ withRange_ TkOpenBrace } -- openBrace }
<0,code> "}"			{ withRange_ TkCloseBrace } -- closeBrace }
<0,code> "["			{ withRange_ TkOpenSquare }
<0,code> "]"			{ withRange_ TkCloseSquare }
<0,code> "$"			{ withRange_ TkDollar }
<0,code> @operator		{ withRange (curry TkOp) }
<0,code> \"			{ lex_atom }

{
-- (Fix highlight problems) "
data Token  = TkLambda Range	| TkArrow Range  | TkId (Range, String)
	    | TkLParen Range	| TkRParen Range | TkSemi Range | TkComma Range
	    | TkUniverse (Range,Int)
	    | TkSet Range	| TkType Range	 | TkProp Range | TkISet Range
	    | TkDo Range | TkFrom Range
	    | TkInfix Range	| TkInfixL Range | TkInfixR Range
            | TkBackQuote Range	| TkWhere Range	 | TkData Range
	    | TkInt (Range, Int)| TkDblArrow Range  | TkWith Range
	    | TkOpenBrace Range	| TkCloseBrace Range	| TkRewrite Range
	    | TkOpenSquare Range| TkCloseSquare Range
	    | TkVOpenBrace Range| TkVCloseBrace Range	| TkVSemi Range
	    | TkEqual Range	| TkColon Range    | TkDot Range
	    | TkLet Range	| TkIn Range	    | TkPlugin (Range,String)
	    | TkSig Range	| TkStruct Range
	    | TkMutual Range	| TkClass Range	    | TkInstance Range
	    | TkExtends Range	| TkDollar Range    | TkAbstract Range
            | TkUnderscore Range| TkOp (Range, String)
	    | TkTeX String	| TkAtom (Range,String)
	    | TkEOF
    deriving Show

scanToken :: Parser Token
scanToken =
    do	inp <- getLexInput
	ls:_ <- getLexState
	case alexScan (foolAlex inp) ls of
	    AlexEOF			-> scanEOF inp TkEOF
	    AlexError _			-> scanError inp
	    AlexSkip inp' len		-> scanSkip (newInput inp inp' len)
						    scanToken
	    AlexToken inp' len action	-> action inp (newInput inp inp' len) len
		

lexer k = scanToken >>= k

newInput (_,s,_) (pn,_,_) len =
    case drop (len - 1) s of
	c:s'	-> (pn, s', c)
	_	-> error "impossible: Syntax.Lexer.newInput"

foolAlex (pn,s,c) = (pn, map fool s, c)
    where
	fool c
	    | isUnicodeId c = 'z'
	    | isUnicodeOp c = '+'
	    | otherwise	    = c

-- States

begin :: LexState -> LexAction Token
begin code _ _ _ =
    do	pushLexState code
	scanToken

begin_ :: LexState -> LexAction Token
begin_ code _ inp' _ =
    do	pushLexState code
	scanSkip inp' scanToken

end_ :: LexAction Token
end_ _ inp' _ =
    do	popLexState
	scanSkip inp' scanToken

popAnd :: LexAction a -> LexAction a
popAnd a inp inp' n =
    do	popLexState
	a inp inp' n

-- Layout

openBrace :: LexAction Token
openBrace = scanTok $ \_ ->
    do	pushContext NoLayout
	r <- getParseRange
	return (TkOpenBrace r)

closeBrace :: LexAction Token
closeBrace = scanTok $ \_ ->
    do	popContext
	r <- getParseRange
	return (TkCloseBrace r)

beginningOfLine :: LexAction Token
beginningOfLine (p,_,_) _ _ =
    do	offs <- getOffside p
	case offs of
	    LT	-> do	popContext
			return (TkVCloseBrace (Range p p))
	    EQ	-> do	popLexState
			return (TkVSemi (Range p p))
	    GT	-> do	popLexState
			scanToken

layoutLeft :: LexAction Token
layoutLeft (p,_,_) _ _ =
    do	popLexState
	pushLexState bol
	return (TkVCloseBrace (Range p p))

newLayoutContext :: LexAction Token
newLayoutContext (p,_,_) _ _ =
    do	popLexState
	let offset = posCol p
	ctx <- getContext
	case ctx of
	    Layout prev_off : _ | prev_off >= offset ->
		do  -- token is indented to the left of the previous context.
		    -- we must generate a {} sequence now.
		    pushLexState layout_left
		    return (TkVOpenBrace (Range p p))
	    _ -> do setContext (Layout offset : ctx)
		    return (TkVOpenBrace (Range p p))

notFollowedBy :: Char -> AlexInput -> AlexInput -> Int -> AlexInput -> Bool
notFollowedBy c _ _ _ (_,i,_) =
    null i || head i /= c

atEOL _ _ _ (_,i,c) = null i || c == '\n'

isBeginCode _ _ _ (_,i,_) = null i || ("\\begin{code}" `isPrefixOf` i)

withLayout :: LexAction r -> LexAction r
withLayout a i1 i2 n =
    do	pushLexState layout
	a i1 i2 n

-- Strings and atoms

lit_error = parseError "lexical error in string literal"

lex_atom :: LexAction Token
lex_atom = lex_string_tok '"' (curry TkAtom)

lex_string_tok :: Char -> (Range -> String -> a) -> LexAction a
lex_string_tok del mkTok (p,_,_) i2 n =
    do	setLexInput i2
	setLastPos p
	tok <- lex_string del ""
	r <- getParseRange
	return $ mkTok r tok

lex_string :: Char -> String -> Parser String
lex_string del s =
    do	i <- getLexInput
	case alexGetChar i of
	    Nothing -> lit_error

	    Just (c,i) | c == del  ->
		do  setLexInput i
		    return (reverse s)

	    Just ('\\',i)
		| Just ('&',i) <- next ->
		    do setLexInput i
		       lex_string del s
		| Just (c,i) <- next, isSpace c ->
		    do setLexInput i
		       lex_stringgap del s
		where next = alexGetChar i

	    Just _ ->
		do c <- lex_char
		   lex_string del (c:s)

lex_stringgap del s =
    do	c <- getCharOrFail
	case c of
	    '\\'	    -> lex_string del s
	    c | isSpace c   -> lex_stringgap del s
	    _other	    -> lit_error

lex_char :: Parser Char
lex_char =
    do	mc <- getCharOrFail
	case mc of
	    '\\'    -> lex_escape
	    c	    -> return c

lex_escape :: Parser Char
lex_escape =
    do	c <- getCharOrFail
	case c of
	    'a'   -> return '\a'
	    'b'   -> return '\b'
	    'f'   -> return '\f'
	    'n'   -> return '\n'
	    'r'   -> return '\r'
	    't'   -> return '\t'
	    'v'   -> return '\v'
	    '\\'  -> return '\\'
	    '"'   -> return '\"'
	    '\''  -> return '\''
	    '^'   -> do c <- getCharOrFail
			if c >= '@' && c <= '_'
			    then return (chr (ord c - ord '@'))
			    else lit_error

	    'x'   -> readNum isHexDigit 16 hexDigit
	    'o'   -> readNum isOctDigit  8 octDecDigit
	    x | isDigit x -> readNum2 isDigit 10 octDecDigit (octDecDigit x)

	    c1 ->
		do  i <- getLexInput
		    case alexGetChar i of
			Nothing		-> lit_error
			Just (c2,i2)	-> 
			    case alexGetChar i2 of
				Nothing		-> lit_error
				Just (c3,i3)	-> 
				   let str = [c1,c2,c3] in
				   case [ (c,rest) | (p,c) <- silly_escape_chars,
						     Just rest <- [maybePrefixMatch p str] ] of
					  (escape_char,[]):_ -> do
						setLexInput i3
						return escape_char
					  (escape_char,_:_):_ -> do
						setLexInput i2
						return escape_char
					  [] -> lit_error

maybePrefixMatch :: String -> String -> Maybe String
maybePrefixMatch []    rest = Just rest
maybePrefixMatch (_:_) []   = Nothing
maybePrefixMatch (p:pat) (r:rest)
  | p == r    = maybePrefixMatch pat rest
  | otherwise = Nothing

hexDigit :: Char -> Int
hexDigit c | isDigit c = ord c - ord '0'
           | otherwise  = ord (toLower c) - ord 'a' + 10

octDecDigit :: Char -> Int
octDecDigit c = ord c - ord '0'

readNum :: (Char -> Bool) -> Int -> (Char -> Int) -> Parser Char
readNum is_digit base conv =
    do	c <- getCharOrFail
	if is_digit c 
	    then readNum2 is_digit base conv (conv c)
	    else lit_error

readNum2 is_digit base conv i =
    do	input <- getLexInput
	read i input
    where
	read i input =
	    case alexGetChar input of
		Just (c,input') | is_digit c ->
		    read (i*base + conv c) input'
		_other ->
		    do	setLexInput input
			if i >= 0 && i <= 0x10FFFF
			    then return (chr i)
			    else lit_error

silly_escape_chars = [
	("NUL", '\NUL'),
	("SOH", '\SOH'),
	("STX", '\STX'),
	("ETX", '\ETX'),
	("EOT", '\EOT'),
	("ENQ", '\ENQ'),
	("ACK", '\ACK'),
	("BEL", '\BEL'),
	("BS", '\BS'),
	("HT", '\HT'),
	("LF", '\LF'),
	("VT", '\VT'),
	("FF", '\FF'),
	("CR", '\CR'),
	("SO", '\SO'),
	("SI", '\SI'),
	("DLE", '\DLE'),
	("DC1", '\DC1'),
	("DC2", '\DC2'),
	("DC3", '\DC3'),
	("DC4", '\DC4'),
	("NAK", '\NAK'),
	("SYN", '\SYN'),
	("ETB", '\ETB'),
	("CAN", '\CAN'),
	("EM", '\EM'),
	("SUB", '\SUB'),
	("ESC", '\ESC'),
	("FS", '\FS'),
	("GS", '\GS'),
	("RS", '\RS'),
	("US", '\US'),
	("SP", '\SP'),
	("DEL", '\DEL')
	]

getCharOrFail :: Parser Char
getCharOrFail =
    do	i <- getLexInput
	case alexGetChar i of
	    Nothing	-> parseError "unexpected end-of-file in string literal"
	    Just (c,i)	-> do setLexInput i
			      return c

-- Nested comments


-- Nested comments require traversing by hand, they can't be parsed
-- using regular expressions.
nestedComment :: LexAction Token
nestedComment (p,_,_) i1 n =
	go 1 i1
    where
	go 0 input = do	setLexInput input
			scanToken
	go n input =
	    case alexGetChar input of
		Nothing		-> err
		Just (c,input)	->
		    case c of
			'-' -> case alexGetChar input of
			    Nothing		-> err
			    Just ('}', input)	-> go (n-1) input
			    Just (c, _)         -> go n input
			'{' -> case alexGetChar input of
			    Nothing		-> err
			    Just ('-',input')	-> go (n+1) input'
			    Just (c,input)	-> go n input
			c -> go n input

        err = parseErrorAt p "Unterminated `{-'"

}
